@@ -1,69 +1,72 @@
-/* Copyright (C) 1989, 1992, 1993 Aladdin Enterprises.  All rights reserved.
-
-This file is part of Ghostscript.
-
-Ghostscript is distributed in the hope that it will be useful, but
-WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-to anyone for the consequences of using it or for whether it serves any
-particular purpose or works at all, unless he says so in writing.  Refer
-to the Ghostscript General Public License for full details.
-
-Everyone is granted permission to copy, modify and redistribute
-Ghostscript, but only under the conditions described in the Ghostscript
-General Public License.  A copy of this license is supposed to have been
-given to you along with Ghostscript so you can know your rights and
-responsibilities.  It should be in a file named COPYING.  Among other
-things, the copyright notice and this notice must be preserved on all
-copies.  */
+/* Copyright (C) 1989, 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+  
+  This file is part of GNU Ghostscript.
+  
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer
+  to the GNU Ghostscript General Public License for full details.
+  
+*/
 
 /* gdevpcfb.c */
-/* IBM PC EGA and VGA display drivers for Ghostscript */
-/* This is fundamentally an EGA driver with some parameters */
-/* that allow it to drive larger displays. */
+/* IBM PC frame buffer (EGA/VGA) drivers */
 #include "memory_.h"
+#include "gconfigv.h"			/* for USE_ASM */
 #include "gx.h"
 #include "gserrors.h"
+#include "gsparam.h"
 #include "gxdevice.h"
+#include "gdevpccm.h"
 #include "gdevpcfb.h"
 
 /* Macro for casting gx_device argument */
 #define fb_dev ((gx_device_ega *)dev)
 
-/* Define the color spectrum */
-#define black 0
-#define blue 1
-#define green 2
-#define cyan 3
-#define red 4
-#define magenta 5
-#define brown 6
-#define white 7
-#define dgray 8				/* dark gray is not very usable */
-#define lblue 9
-#define lgreen 10
-#define lcyan 11
-#define lred 12
-#define lmagenta 13
-#define yellow 14
-#define bwhite 15
-
 /* Procedure record */
+private dev_proc_map_rgb_color(ega0_map_rgb_color);
+private dev_proc_map_rgb_color(ega1_map_rgb_color);
+#define ega2_map_rgb_color pc_4bit_map_rgb_color
+private dev_proc_map_color_rgb(ega01_map_color_rgb);
+#define ega2_map_color_rgb pc_4bit_map_color_rgb
+#if ega_bits_of_color == 0
+#   define ega_map_rgb_color ega0_map_rgb_color
+#   define ega_map_color_rgb ega01_map_color_rgb
+#else
+# if ega_bits_of_color == 1
+#   define ega_map_rgb_color ega1_map_rgb_color
+#   define ega_map_color_rgb ega01_map_color_rgb
+# else
+#   define ega_map_rgb_color ega2_map_rgb_color
+#   define ega_map_color_rgb ega2_map_color_rgb
+# endif
+#endif
+#define ega_std_procs\
+	ega_open,\
+	NULL,			/* get_initial_matrix */\
+	NULL,			/* sync_output */\
+	NULL,			/* output_page */\
+	ega_close,\
+	ega_map_rgb_color,\
+	ega_map_color_rgb,\
+	ega_fill_rectangle,\
+	ega_tile_rectangle,\
+	ega_copy_mono,\
+	ega_copy_color,\
+	NULL,			/* draw_line */\
+	ega_get_bits
+
 private gx_device_procs ega_procs = {
-	ega_open,
-	gx_default_get_initial_matrix,
-	gx_default_sync_output,
-	gx_default_output_page,
-	ega_close,
-	ega_map_rgb_color,
-	ega_map_color_rgb,
-	ega_fill_rectangle,
-	ega_tile_rectangle,
-	ega_copy_mono,
-	ega_copy_color,
-	gx_default_draw_line,
-	ega_get_bits,
-	gx_default_get_props,
-	gx_default_put_props
+	ega_std_procs
+};
+
+private dev_proc_get_params(svga16_get_params);
+private dev_proc_put_params(svga16_put_params);
+private gx_device_procs svga16_procs = {
+	ega_std_procs,
+	svga16_get_params,
+	svga16_put_params
 };
 
 /* All the known instances */
@@ -73,40 +76,32 @@ gx_device_ega gs_ega_device =
 		/* VGA */
 gx_device_ega gs_vga_device =
 	ega_device("vga", ega_procs, 80, 480, 1.0, 0x12);
-		/* ATI Wonder SuperVGA, 800x600, 16-color mode */
-gx_device_ega gs_atiw16_device =
-	ega_device("atiw16", ega_procs, 100, 600, 1.0, 0x54);
-		/* Trident SuperVGA, 800x600, 16-color mode */
-gx_device_ega gs_tvga16_device =
-	ega_device("tvga16", ega_procs, 100, 600, 1.0, 0x5b);
-		/* Tseng Labs SuperVGA, 800x600, 16-color mode */
-gx_device_ega gs_tseng16_device =
-	ega_device("tseng16", ega_procs, 100, 600, 1.0, 0x29);
+		/* Generic SuperVGA, 800x600, 16-color mode */
+gx_device_ega gs_svga16_device =
+	ega_device("svga16", svga16_procs, 100, 600, 1.0, 0x29 /*Tseng*/);
 
-/* Save the EGA mode */
-private int ega_save_mode = -1;
+/* Save the BIOS state */
+private pcfb_bios_state pcfb_save_state = { -1 };
 
 /* Initialize the EGA for graphics mode */
 int
 ega_open(gx_device *dev)
 {	/* Adjust the device resolution. */
-	/* This is a hack, pending refactoring of the put_props machinery. */
+	/* This is a hack, pending refactoring of the put_params machinery. */
 	switch ( fb_dev->video_mode )
 	{
 	case 0x10:	/* EGA */
 		gx_device_adjust_resolution(dev, 640, 350, 1); break;
 	case 0x12:	/* VGA */
 		gx_device_adjust_resolution(dev, 640, 480, 1); break;
-	case 0x54:      /* ATI Wonder */
-	case 0x5b:	/* Trident */
-	case 0x29:	/* Tseng */
+	default:	/* 800x600 SuperVGA */
 		gx_device_adjust_resolution(dev, 800, 600, 1); break;
 	}
-	if ( ega_save_mode < 0 )
-		ega_save_mode = ega_get_mode();
+	if ( pcfb_save_state.display_mode < 0 )
+		pcfb_get_state(&pcfb_save_state);
 	/* Do implementation-specific initialization */
-	ega_set_signals(dev);
-	ega_set_mode(fb_dev->video_mode);
+	pcfb_set_signals(dev);
+	pcfb_set_mode(fb_dev->video_mode);
 	set_s_map(-1);			/* enable all maps */
 	return 0;
 }
@@ -114,69 +109,81 @@ ega_open(gx_device *dev)
 /* Reinitialize the EGA for text mode */
 int
 ega_close(gx_device *dev)
-{	if ( ega_save_mode >= 0 )
-		ega_set_mode(ega_save_mode);
+{	if ( pcfb_save_state.display_mode >= 0 )
+	  pcfb_set_state(&pcfb_save_state);
+	return 0;
+}
+
+/* Get/put the display mode parameter. */
+int
+svga16_get_params(gx_device *dev, gs_param_list *plist)
+{	int code = gx_default_get_params(dev, plist);
+	if ( code < 0 )
+	  return code;
+	return param_write_int(plist, "DisplayMode", &fb_dev->video_mode);
+}
+int
+svga16_put_params(gx_device *dev, gs_param_list *plist)
+{	int ecode = 0;
+	int code;
+	int imode = fb_dev->video_mode;
+	const char _ds *param_name;
+
+	switch ( code = param_read_int(plist, (param_name = "DisplayMode"), &imode) )
+	  {
+	  default:
+		ecode = code;
+		param_signal_error(plist, param_name, ecode);
+	  case 0:
+	  case 1:
+		break;
+	  }
+
+	if ( ecode < 0 )
+	  return ecode;
+	code = gx_default_put_params(dev, plist);
+	if ( code < 0 )
+	  return code;
+
+	if ( imode != fb_dev->video_mode )
+	  {	if ( dev->is_open )
+		  gs_closedevice(dev);
+		fb_dev->video_mode = imode;
+	  }
+
 	return 0;
 }
 
 /* Map a r-g-b color to an EGA color code. */
-gx_color_index
-ega_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
+#define Nb gx_color_value_bits
+private gx_color_index
+ega0_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
+  gx_color_value b)
+{	return pc_4bit_map_rgb_color(dev, r, r, r);
+}
+private gx_color_index
+ega1_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
   gx_color_value b)
 {
-#define c12 (gx_max_color_value / 2)
-#define c13 (gx_max_color_value / 3)
-#define c23 (gx_max_color_value - c13)
-#define cmono() ((gx_color_index)mono_color[r >> (gx_color_value_bits - 2)])
-#if ega_bits_of_color == 0		/* monochrome */
-	static const byte mono_color[4] = { black, white, white, bwhite };
-	return cmono();
-#else
-# if ega_bits_of_color == 1
-	static const byte mono_color[4] = { black, white, white, bwhite };
-	if ( r == g && g == b )
-		return cmono();
-	return (gx_color_index)
-		((r > c12 ? 4 : 0) |
-		 (g > c12 ? 10 : 0) |
-		 (b > c12 ? 1 : 0));
-# else
-	static const byte g0[3][3] =
-	 {{black,blue,lblue},{red,magenta,lmagenta},{lred,lmagenta,lmagenta}};
-	static const byte g1[3][3] =
-	 {{green,cyan,lcyan},{brown,white,lcyan},{yellow,yellow,lmagenta}};
-	static const byte g2[3][3] =
-	 {{lgreen,lgreen,lcyan},{lgreen,lgreen,lcyan},{yellow,yellow,bwhite}};
-	return (gx_color_index)
-		((g >= c23 ? g2 : g >= c13 ? g1 : g0)
-		 [r >= c23 ? 2 : r >= c13 ? 1 : 0]
-		 [b >= c23 ? 2 : b >= c13 ? 1 : 0]);
-# endif
-#endif
-#undef c12
-#undef c13
-#undef c23
+#define cvtop (gx_color_value)(1 << (Nb - 1))
+	return pc_4bit_map_rgb_color(dev, r & cvtop, g & cvtop, b & cvtop);
 }
+#undef Nb
 
 /* Map a color code to r-g-b. */
-int
-ega_map_color_rgb(gx_device *dev, gx_color_index color,
+#define icolor (int)color
+private int
+ega01_map_color_rgb(gx_device *dev, gx_color_index color,
   gx_color_value prgb[3])
 {
-#define icolor (int)color
-#if rgb_max > 1
-	gx_color_value one =
-		(icolor & 8 ? gx_max_color_value : gx_max_color_value / 3);
-#else
-#  define one (gx_max_color_value / 2 + 1)
-#endif
+#define one (gx_max_color_value / 2 + 1)
 	prgb[0] = (icolor & 4 ? one : 0);
 	prgb[1] = (icolor & 2 ? one : 0);
 	prgb[2] = (icolor & 1 ? one : 0);
 	return 0;
 #undef one
-#undef icolor
 }
+#undef icolor
 
 /* ------ Internal routines ------ */
 
@@ -533,31 +540,42 @@ ega_copy_color(gx_device *dev,
   const byte *base, int sourcex, int raster, gx_bitmap_id id,
   int x, int y, int w, int h)
 {	const byte *line = base + (sourcex >> 1);
-	unsigned lmask = 0x80 >> (x & 7);
-	int bitx = sourcex & 1;
-	int ex = w + bitx;
-	fb_ptr fbline;
+	unsigned mask = 0x80 >> (x & 7);
+	int px = sourcex & 1;
+	fb_ptr fb_line;
+	int fb_raster = fb_dev->raster;
 	fit_copy(dev, base, sourcex, raster, id, x, y, w, h);
-	fbline = mk_fb_ptr(x, y);
+	fb_line = mk_fb_ptr(x, y);
 	set_g_mode(gm_FILL);
 	select_g_mask();
-	while ( --h >= 0 )
+	for ( ; ; px++ )
 	{	const byte *bptr = line;
-		int px = bitx;
-		/* Turbo C will put an unsigned in a register, */
-		/* but not a byte! */
-		register unsigned mask = lmask;
-		fb_ptr fbptr = fbline;
-		do
-		   {	byte_discard(*fbptr);	/* latch frame buffer data */
-			out_g_mask(mask);
-			*fbptr = (px & 1 ? *bptr++ & 0xf : *bptr >> 4);
-			if ( (mask >>= 1) == 0 )
-				mask = 0x80, fbptr++;
-		   }
-		while ( ++px < ex );
-		line += raster;
-		fbline += fb_dev->raster;
+		fb_ptr fbptr = fb_line;
+		int py = h;
+		out_g_mask(mask);
+		if ( px & 1 )
+		{	do
+			   {	byte_discard(*fbptr);	/* latch frame buffer data */
+				*fbptr = *bptr;
+				bptr += raster;
+				fbptr += fb_raster;
+			   }
+			while ( --py );
+			line++;
+		}
+		else
+		{	do
+			   {	byte_discard(*fbptr);	/* latch frame buffer data */
+				*fbptr = *bptr >> 4;
+				bptr += raster;
+				fbptr += fb_raster;
+			   }
+			while ( --py );
+		}
+		if ( !--w )
+			break;
+		if ( (mask >>= 1) == 0 )
+			mask = 0x80, fb_line++;
 	}
 	set_g_mode(gm_DATA);
 	dot_end();
@@ -586,7 +604,7 @@ ega_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
 /* i.e. neither one can be gx_no_color_index (transparent): */
 /* a transparent color means that the tile is colored, not a mask. */
 int
-ega_tile_rectangle(gx_device *dev, const gx_bitmap *tile,
+ega_tile_rectangle(gx_device *dev, const gx_tile_bitmap *tile,
   int x, int y, int w, int h, gx_color_index czero, gx_color_index cone,
   int px, int py)
 #define zero (int)czero
